/* Artur Marcin Laskowski, 8EIBE */

/* Biblioteka STL, algorytmy sortujce, funkcja 'partial_sort'. */

/* Nagwek funkcji

    template <class RandomAccessIterator>
        void partial_sort(RandomAccessIterator first,
                          RandomAccessIterator middle,
                          RandomAccessIterator last);

    template <class RandomAccessIterator, class Compare>
        void partial_sort(RandomAccessIterator first,
                          RandomAccessIterator middle,
                          RandomAccessIterator last,
                          Compare comp);
*/

/* Przemieszcza elementy w kolekcji w zakresie [first, last) tak, e w zakresie
[first, middle) zawiera najmniejsze elementy z caej kolekcji, a w zakresie
[middle, last) znajduj si pozostae, nieuporzdkowane elementy.
   Elementy s porwnywane za pomoc operatora 'operator<' w pierwszej wersji,
bd za pomoc funkcji porwnujcej/funktora w wersji drugiej.
   Do sortowania wykorzystano algorytm 'heapsort', a zoono obliczeniow
funkcji szacuje si wg wzoru:

    O = (last - first) * log(middle - first)
*/

#include    <algorithm>
#include    <iostream>
#include    <vector>

using namespace std;

/* Rodzaj sortowania */
enum    SortBy {byName, byAge, byWeight};

/* Przykadowa klasa suca prezentacji dziaania algorytmu 'partial_sort'.
Implementuje uproszczon kart pacjenta, ktra przechowuje nazwisko, wiek i
wag pacjenta. Statyczne pole 'sort' przechowuje biec metod sortowania.
Klasa dostarcza operatora 'operator<', ktry porwnuje odpowiednie pola dwch
obiektw klasy. */
class   PatientCard {
    private:
        string          name;                   //Nazwisko
        int             age;                    //Wiek
        double          weight;                 //Waga

        static SortBy   sort;                   //Rodzaj sortowania
    public:
        /* Konstruktor zerowy. */
        PatientCard() : name("name"), age(0), weight(0.0) {}
        /* Konstruktor inicjujcy. */
        PatientCard(const string& _name, const int& _age, const double& _weight) :
            name(_name), age(_age), weight(_weight) {}

        const string&   getName() const {       //Zwraca nazwisko
            return name;
        }

        const int       getAge() const {        //Zwraca wiek
            return age;
        }

        const double    getWeight() const {     //Zwraca wag
            return weight;
        }

        void    setSort(SortBy _sort) {         //Ustala rodzaj sortowania
            sort    = _sort;
        }

        /* Operator mniejszoci, sucy do porwnywania w algorytmie. Porwnuje
        odpowiednie pola dwch obiektw, zwraca 'true', jeli s uporzdkowane
        rosnco, i 'false', jeli nie. Do wyboru pl do porwnywania suy pole
        'sort'. */
        bool    operator<(const PatientCard& _card) {
            switch (sort) {
                case    byName:
                    if (name.compare(_card.name) < 0)
                        return true;
                    break;
                case    byAge:
                    if (age < _card.age)
                        return true;
                    break;
                case    byWeight:
                    if (weight < _card.weight)
                        return true;
                    break;
                default:
                    return false;
            }

            return false;
        };

        /* Zaprzyjaniony operator, pozwalajcy na wysanie danych obiektu do
        strumienia wyjciowego. */
        friend ostream& operator<<(ostream& _cout, const PatientCard& _card);
};

SortBy  PatientCard::sort   = byName;           //Inicjacja statycznego pola 'sort'

/* Zaprzyjaniony operator. */
ostream&    operator<<(ostream& _cout, const PatientCard& _card) {
    _cout << _card.name << "\n\t\twiek: " << _card.age << "\twaga: " << _card.weight << endl;

    return _cout;
}

/* Funkcja porwnujca dwa obiekty klasy 'PatientCard'. W tym przykadzie
sortuje elementy wg malejcej wagi. */
bool    compareCard(const PatientCard& _a, const PatientCard& _b) {
    if (_a.getWeight() > _b.getWeight())
        return true;
    return false;
}

/* Funktor - klasa 'opakowujca' operator 'operator()', w ktrym zdefiniowano
porwnywanie dwch obiektw klasy 'PatientCard'. Tutaj umoliwia sortowanie
wg malejcego wieku. */
class   functorCard {
    public:
        bool    operator()(const PatientCard& _a, const PatientCard& _b) {
            if (_a.getAge() > _b.getAge())
                return true;

            return false;
        }
};

/* Funkcja gwna programu, uruchamia ca prezentacj algorytmu. */
int     main() {
    vector<int>             first;              //Kolekcja liczb cakowitych
    vector<int>::iterator   iter;               //Iterator do tej kolekcji

    /* Wypenianie kolekcji przykladowymi liczbami */
    first.push_back(23);
    first.push_back(16);
    first.push_back(54);
    first.push_back(21);
    first.push_back(14);
    first.push_back(38);
    first.push_back(63);
    first.push_back(29);

    cout << "Przyklad pierwszy\n\n";
    cout << "Liczby calkowite 'int':\n\n";

    /* Wywietlenie zawartoci kolekcji */
    for (iter = first.begin(); iter != first.end(); ++iter)
        cout << *iter << " ";

    /* Sortowanie kolekcji, do uzyskania 4 posortowanych elementw */
    partial_sort(first.begin(), first.begin() + 4, first.end());

    cout << "\n\nDane czesciowo posortowane - pierwsze 4 elementy:\n\n";

    /* Wywietlenie zawartoci czciowo posortowanej kolekcji */
    for (iter = first.begin(); iter != first.end(); ++iter)
        cout << *iter << " ";

    vector<PatientCard>             second;     //Kolekcja kart pacjenta
    vector<PatientCard>::iterator   iter2;      //Iterator do tej kolekcji

    /* Wypenienie kolekcji przykadowymi danymi */
    second.push_back(PatientCard("Wilson", 32, 78.5));
    second.push_back(PatientCard("Pearson", 27, 63.2));
    second.push_back(PatientCard("Kingsley", 43, 89.0));
    second.push_back(PatientCard("Smith", 18, 57.4));
    second.push_back(PatientCard("Wiliams", 22, 82.1));
    second.push_back(PatientCard("Crouch", 37, 102.6));
    second.push_back(PatientCard("Michelson", 64, 96.4));
    second.push_back(PatientCard("Bright", 54, 60.5));

    vector<PatientCard> third   = second;       //Kolejne kolekcje
    vector<PatientCard> fourth  = second;

    cout << "\n\n\nPrzyklad drugi\n\n";
    cout << "Klasa 'PatientCard', wbudowany operator 'operator<':\n\n";

    for (iter2 = second.begin(); iter2 != second.end(); ++iter2)
        cout << *iter2;

    /* Sortowanie wg wbudowanego operatora 'operator<' do uzyskania 3 elementw
    posortowanych */
    partial_sort(second.begin(), second.begin() + 3, second.end());

    cout << "\nDane czesciowo posortowane - pierwsze 3, rosnaco alfabetycznie:\n\n";

    for (iter2 = second.begin(); iter2 != second.end(); ++iter2)
        cout << *iter2;

    cout << "\n\n\nPrzyklad trzeci\n\n";
    cout << "Klasa 'PatientCard', funkcja porownujaca:\n";

    /* Sortowanie za pomoc funcji porwnujcej, do uzyskania 4 elementw
    posortowanych. */
    partial_sort(third.begin(), third.begin() + 4, third.end(), compareCard);

    cout << "Dane czesciowo posortowane - pierwsze 4, malejaco wg wagi:\n\n";

    for (iter2 = third.begin(); iter2 != third.end(); ++iter2)
        cout << *iter2;

    cout << "\n\n\nPrzyklad czwarty\n\n";
    cout << "Klasa 'PatientCard', funktor:\n";

    /* Sortowaniw za pomoc funktora, do uzyskania 5 posortowanych elementw. */
    partial_sort(fourth.begin(), fourth.begin() + 5, fourth.end(), functorCard());

    cout << "Dane czesciowo posortowane - pierwsze 5, malejaco wg wieku:\n\n";

    for (iter2 = fourth.begin(); iter2 != fourth.end(); ++iter2)
        cout << *iter2;

    return 0;
}
